# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.logger
    -----------------------

    Logging setup.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import unicode_literals
import logging
from .compat import NullHandler


LOGGER = logging.getLogger('pycampbellcr1000')
LOGGER.addHandler(NullHandler())


def active_logger():
    '''Initialize a speaking logger with stream handler (stderr).'''
    LOGGER = logging.getLogger('pycampbellcr1000')

    LOGGER.setLevel(logging.INFO)

    # Default to logging to stderr.
    formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s ')
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)

    LOGGER.addHandler(stream_handler)



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.main
    ---------------------

    The public API and command-line interface to PyCampbellCR1000.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
import os
import argparse

from datetime import datetime

# Make sure the logger is configured early:
from . import VERSION
from .logger import active_logger
from .compat import stdout
from .device import CR1000
from .utils import csv_to_dict, ListDict


NOW = datetime.now().strftime("%Y-%m-%d %H:%M")


def gettime_cmd(args, device):
    '''Gettime command.'''
    print("%s" % (device.gettime()))


def settime_cmd(args, device):
    '''Settime command.'''
    old_time = device.gettime()
    device.settime(datetime.strptime(args.datetime, "%Y-%m-%d %H:%M"))
    print("Old Time : %s" % old_time.strftime("%Y-%m-%d %H:%M"))
    print("Current Time : %s" % device.gettime().strftime("%Y-%m-%d %H:%M"))


def getprogstat_cmd(args, device):
    '''Getprogstat command.'''
    data = device.getprogstat()
    for key in data.keys():
        print("%s : %s" % (key, ("%s" % data[key]).strip('\r\n')))


def getsettings_cmd(args, device):
    '''Getsettings command.'''
    args.delim = args.delim.decode("string-escape")
    data = device.settings
    for item in data:
        item["SettingValue"] = repr(item["SettingValue"])
    args.output.write("%s" % data.to_csv(delimiter=args.delim))


def listfiles_cmd(args, device):
    '''Listfiles command.'''
    for filename in device.list_files():
        print(filename.decode('utf-8'))


def getfile_cmd(args, device):
    '''Getfile command.'''
    args.output.write("%s" % device.getfile(args.filename.decode('utf-8')))


def listtables_cmd(args, device):
    '''Listtables command.'''
    for tablename in device.list_tables():
        print(tablename.decode('utf-8'))


def getdata_cmd(args, device, header=True, exclude_first=False):
    '''Getdata command.'''
    args.delim = args.delim.decode("string-escape")
    if args.start is not None:
        args.start = datetime.strptime(args.start, "%Y-%m-%d %H:%M")
    if args.stop is not None:
        args.stop = datetime.strptime(args.stop, "%Y-%m-%d %H:%M")
    print("Your download is starting.")
    total_records = 0
    generator = device.get_data_generator(args.table, args.start, args.stop)
    for i, records in enumerate(generator):
        if exclude_first:
            records = ListDict(records[1:])
            exclude_first = False
        total_records += len(records)
        print("Packet %d with %d records" % (i, len(records)))
        args.output.write("%s" % records.to_csv(delimiter=args.delim,
                                                header=header))
        if header:
            header = False

    print("---------------------------")
    if total_records == 0:
        print("No new records were found﻿")
    elif total_records == 1:
        print("1 new record was found")
    else:
        print("%d new records were found" % total_records)


def update_cmd(args, device):
    '''Update command.'''
    # create file if not exist
    with file(args.db, 'a'):
        os.utime(args.db, None)
    with open(args.db, 'r') as file_db:
        db = csv_to_dict(file_db, delimiter=args.delim)
    with open(args.db, 'a') as file_db:
        args.start = None
        args.stop = None
        if len(db) > 0:
            db = db.sorted_by("Datetime", reverse=True)
            format = "%Y-%m-%d %H:%M:%S"
            start_date = datetime.strptime(db[0]['Datetime'], format)
            # exclude this record
            args.start = start_date.strftime("%Y-%m-%d %H:%M")
            args.output = file_db
            getdata_cmd(args, device, header=False, exclude_first=True)
        else:
            args.output = file_db
            getdata_cmd(args, device, header=True)


def get_cmd_parser(cmd, subparsers, help, func):
    '''Make a subparser command.'''
    formatter_class = argparse.ArgumentDefaultsHelpFormatter
    parser = subparsers.add_parser(cmd, help=help, description=help,
                                   formatter_class=formatter_class)
    parser.add_argument('--timeout', default=10.0, type=float,
                        help="Connection link timeout")
    parser.add_argument('--src', default=0x802, type=int,
                        help='Source node ID')
    parser.add_argument('--dest', default=0x001, type=int,
                        help='Destination node ID')
    parser.add_argument('--code', default=0x0000, type=int,
                        help='Datalogger security code')
    parser.add_argument('--debug', action="store_true", default=False,
                        help='Display log')
    parser.add_argument('url', action="store",
                        help="Specifiy URL for connection link. "
                        "E.g. tcp:iphost:port "
                        "or serial:/dev/ttyUSB0:19200:8N1")
    parser.set_defaults(func=func)
    return parser


def main():
    '''Parse command-line arguments and execute CR1000 command.'''

    formatter_class = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(prog='pycr1000',
                                     description='Communication tools for '
                                                 'Campbell CR1000-type '
                                                 'Datalogger',
                                     formatter_class=formatter_class)

    parser.add_argument('--version', action='version',
                        version='PyCR1000 version %s' % VERSION,
                        help='Print PyCR1000’s version number and exit.')

    subparsers = parser.add_subparsers(title='The PyCR1000 commands')
    # gettime command
    subparser = get_cmd_parser('gettime', subparsers,
                               help='Print the current datetime of the'
                                    ' datalogger.',
                               func=gettime_cmd)

    # settime command
    subparser = get_cmd_parser('settime', subparsers,
                               help='Set the given datetime argument on the'
                                    ' datalogger.',
                               func=settime_cmd)
    subparser.add_argument('datetime', help='The chosen datetime value. '
                                            '(like : "%s")' % NOW)

    # getprogstat command
    subparser = get_cmd_parser('getprogstat', subparsers,
                               help='Retrieve available programming '
                                    'statistics information from the '
                                    'datalogger.',
                               func=getprogstat_cmd)

    # getsettings command
    subparser = get_cmd_parser('getsettings', subparsers,
                               help='Retrieve the datalogger settings.',
                               func=getsettings_cmd)
    subparser.add_argument('--output', action='store', default=stdout,
                           type=argparse.FileType('w', 0),
                           help='Filename where output is written')
    subparser.add_argument('--delim', action='store', default=",",
                           help='CSV char delimiter')

    # listfiles command
    subparser = get_cmd_parser('listfiles', subparsers,
                               help='List all files stored in the datalogger.',
                               func=listfiles_cmd)

    # getfile command
    subparser = get_cmd_parser('getfile', subparsers,
                               help='Get the file content from the '
                                    'datalogger.',
                               func=getfile_cmd)
    subparser.add_argument('filename', action="store",
                           help="Filename to be downloaded.")
    subparser.add_argument('output', action='store',
                           type=argparse.FileType('w', 0),
                           help='Filename where output is written')

    # listtables command
    subparser = get_cmd_parser('listtables', subparsers,
                               help='List all tables stored in '
                                    'the datalogger.',
                               func=listtables_cmd)

    # getdata command
    subparser = get_cmd_parser('getdata', subparsers,
                               help='Extract data from the datalogger '
                                    'between start datetime and stop datetime.'
                                    'By default the entire contents will be '
                                    'downloaded.',
                               func=getdata_cmd)
    subparser.add_argument('table', action="store",
                           help="The table name used for data collection")
    subparser.add_argument('output', action='store',
                           type=argparse.FileType('w', 0),
                           help='Filename where output is written')
    subparser.add_argument('--start', help='The beginning datetime record '
                                           '(like : "%s")' % NOW)
    subparser.add_argument('--stop', help='The stopping datetime record '
                                          '(like : "%s")' % NOW)
    subparser.add_argument('--delim', action='store', default=",",
                           help='CSV char delimiter')

    # update command
    subparser = get_cmd_parser('update', subparsers,
                               help='Update CSV database records by getting '
                               'automatically new records.',
                               func=update_cmd)
    subparser.add_argument('table', action="store",
                           help="The table name used for data collection")
    subparser.add_argument('--delim', action="store", default=",",
                           help='CSV char delimiter')
    subparser.add_argument('db', action="store", help='The CSV file database')

    # Parse argv arguments
    args = parser.parse_args()

    if args.debug:
        active_logger()
        device = CR1000.from_url(args.url, args.timeout, args.dest, args.src,
                                 args.code)
        args.func(args, device)
    else:
        try:
            device = CR1000.from_url(args.url, args.timeout, args.dest,
                                     args.src, args.code)
            args.func(args, device)
        except Exception as e:
            parser.error('%s' % e)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000
    ----------------

    The public API and command-line interface to PyCampbellCR1000.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
# Make sure the logger is configured early:
from .logger import LOGGER, active_logger
from .pakbus import PakBus
from .device import CR1000


VERSION = '0.2dev'
__version__ = VERSION



# coding: utf8
'''
    PyCampbellCRX.tests.test_device
    -------------------------------

    The device test suite for PyCampbelCR1000.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import unicode_literals
from datetime import datetime, timedelta

from ..logger import active_logger
from ..device import CR1000


active_logger()


class TestDevice:
    ''' Test device.'''

    def setup_class(self):
        '''Setup common data.'''
        pass

    def test_gettime(self, url):
        device = CR1000.from_url(url, 1)
        assert isinstance(device.gettime(), datetime)

    def test_settime(self, url):
        device = CR1000.from_url(url, 1)
        now = datetime.now().replace(second=10, microsecond=0)
        device_time = device.settime(now)
        assert device_time.second in (10, 11, 12)
        assert now == device_time.replace(second=10)
        lastday = now - timedelta(days=1)
        tomorrow = now + timedelta(days=1)
        assert lastday == device.settime(lastday).replace(second=10)
        assert tomorrow == device.settime(tomorrow).replace(second=10)
        device.settime(datetime.now())

    def test_settings(self, url):
        device = CR1000.from_url(url, 1)
        assert device.settings[0]['SettingId'] == 0
        print (device.settings[0]['SettingValue'])
        print (type(device.settings[0]['SettingValue']))
        assert b"CR" in device.settings[0]['SettingValue']
        assert device.settings[0]['ReadOnly'] == 1
        assert device.settings[0]['LargeValue'] == 0

    def test_list_and_get_file(self, url):
        device = CR1000.from_url(url, 1)
        files = device.list_files()
        if len(files) > 1:
            for filename in files[1:]:
                filename = filename.decode('utf-8')
                if filename != "CPU:":
                    fd = device.getfile(filename)
                    assert fd != b""
                    break

    def get_data_generator(self, device, table):
        generator = device.get_data_generator(table, None, None)
        try:
            first_records = next(generator)
        except StopIteration:
            first_records = []
        if len(first_records) > 0:
            rec = first_records[0]
            for next_rec in first_records[1:]:
                assert rec['Datetime'] < next_rec['Datetime']
                rec = next_rec

    def get_data(self, device, table):
        start_date = datetime.now() - timedelta(days=2)
        stop_date = datetime.now() - timedelta(days=1, hours=23)
        records = device.get_data(table, start_date, stop_date)
        if len(records) > 0:
            for item in records:
                assert start_date <= item["Datetime"] < stop_date

    def test_list_get_tables(self, url):
        device = CR1000.from_url(url, 1)
        tables = device.list_tables()
        if len(tables) > 0:
            for table in tables:
                table = table.decode('utf-8')
                self.get_data(device, table)
                self.get_data_generator(device, table)

    def test_getprogstat(self, url):
        device = CR1000.from_url(url, 1)
        osversion = device.getprogstat()['OSVer'].decode('utf-8')
        osversion = osversion.split(".")[0]
        assert (osversion == "CR800" or osversion == "CR1000")



import pytest


def pytest_addoption(parser):
    parser.addoption("--url", dest="url", help="PyLink url connection.")

def pytest_generate_tests(metafunc):
    if 'url' in metafunc.funcargnames:
        if not metafunc.config.option.url:
            pytest.skip("test requires url connection")
        else:
            metafunc.parametrize("url", [metafunc.config.option.url])



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.exceptions
    ---------------------------

    Exceptions

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import division, unicode_literals


class NoDeviceException(Exception):
    '''Can not access to device.'''
    def __str__(self):
        return self.__doc__


class BadSignatureException(Exception):
    '''No valid signature.'''
    def __str__(self):
        return self.__doc__


class BadDataException(Exception):
    '''No valid data packet.'''
    def __str__(self):
        return self.__doc__


class DeliveryFailureException(Exception):
    '''Delivery failure.'''
    def __str__(self):
        return self.__doc__



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.pakbus
    -----------------------

    PakBus protocol Implementation.

    Original Author: Dietrich Feist, Max Planck Institute for Biogeochemistry,
                     Jena Germany (PyPak)
    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import division, unicode_literals

import struct
import time

from .compat import ord, chr, is_text, is_py3, bytes
from .logger import LOGGER
from .utils import Singleton
from .exceptions import DeliveryFailureException
from .utils import bytes_to_hex, nsec_to_time


class Transaction(Singleton):
    id = 0

    def next_id(self):
        self.id += 1
        self.id &= 0xFF
        return self.id


class PakBus(object):
    '''Inteface for a pakbus client.

    :param link: A `PyLink` connection.
    :param dest_node: Destination node ID (12-bit int) (default 0x001)
    :param src_node: Source node ID (12-bit int) (default 0x802)
    :param security_code: 16-bit security code (default 0x0000)
    '''

    DATATYPE = {
        'Byte':     {'code':  1, 'fmt': 'B',   'size': 1},
        'UInt2':    {'code':  2, 'fmt': '>H',  'size': 2},
        'UInt4':    {'code':  3, 'fmt': '>L',  'size': 4},
        'Int1':     {'code':  4, 'fmt': 'b',   'size': 1},
        'Int2':     {'code':  5, 'fmt': '>h',  'size': 2},
        'Int4':     {'code':  6, 'fmt': '>l',  'size': 4},
        'FP2':      {'code':  7, 'fmt': '>H',  'size': 2},
        'FP3':      {'code': 15, 'fmt': '3c',  'size': 3},
        'FP4':      {'code':  8, 'fmt': '4c',  'size': 4},
        'IEEE4B':   {'code':  9, 'fmt': '>f',  'size': 4},
        'IEEE8B':   {'code': 18, 'fmt': '>d',  'size': 8},
        'Bool8':    {'code': 17, 'fmt': 'B',   'size': 1},
        'Bool':     {'code': 10, 'fmt': 'B',   'size': 1},
        'Bool2':    {'code': 27, 'fmt': '>H',  'size': 2},
        'Bool4':    {'code': 28, 'fmt': '>L',  'size': 4},
        'Sec':      {'code': 12, 'fmt': '>l',  'size': 4},
        'USec':     {'code': 13, 'fmt': '6c',  'size': 6},
        'NSec':     {'code': 14, 'fmt': '>2l', 'size': 8},
        'ASCII':    {'code': 11, 'fmt': 's',   'size': None},
        'ASCIIZ':   {'code': 16, 'fmt': 's',   'size': None},
        'Short':    {'code': 19, 'fmt': '<h',  'size': 2},
        'Long':     {'code': 20, 'fmt': '<l',  'size': 4},
        'UShort':   {'code': 21, 'fmt': '<H',  'size': 2},
        'ULong':    {'code': 22, 'fmt': '<L',  'size': 4},
        'IEEE4L':   {'code': 24, 'fmt': '<f',  'size': 4},
        'IEEE8L':   {'code': 25, 'fmt': '<d',  'size': 8},
        'SecNano':  {'code': 23, 'fmt': '<2l', 'size': 8},
    }

    # link state
    RING = 0x9
    READY = 0xA
    FINISHED = 0xB

    def __init__(self, link, dest_node=0x001, src_node=0x802,
                 security_code=0x0000):
        self.link = link
        self.src_node = src_node
        self.dest_node = dest_node
        self.security_code = security_code
        self.transaction = Transaction()
        LOGGER.info('Get the node attention')
        self.link.write(b'\xBD\xBD\xBD\xBD\xBD\xBD')

    def write(self, packet):
        '''Send packet over PakBus.'''
        LOGGER.info('Packet data: %s' % bytes_to_hex(packet))
        LOGGER.info('Calculate signature for packet')
        sign = self.compute_signature(packet)
        LOGGER.info('Calculate signature nullifier to create packet')
        nullifier = self.compute_signature_nullifier(sign)
        frame = self.quote(b''.join((packet, nullifier)))
        packet = b''.join((b'\xBD', frame, b'\xBD'))
        LOGGER.info('Write: %s' % bytes_to_hex(packet))
        self.link.write(packet)

    def read(self):
        '''Receive packet over PakBus.'''
        all_bytes = []
        byte = None
        begin = time.time()
        while byte != b'\xBD':
            if time.time() - begin > self.link.timeout:
                return None
            # Read until first \xBD frame character
            byte = self._read_one_byte()
        while byte == b'\xBD':
            # Read unitl first character other than \xBD
            byte = self._read_one_byte()
        while byte != b'\xBD':
            # Read until next occurence of \xBD character
            all_bytes.append(byte)
            byte = self._read_one_byte()

        # Unquote quoted characters
        packet = b''.join(all_bytes)
        LOGGER.info('Read packet: %s' % bytes_to_hex(packet))
        packet = self.unquote(packet)

        # Calculate signature (should be zero)
        if self.compute_signature(packet):
            LOGGER.error('Check signature : Error')
            return None
        else:
            LOGGER.info('Check signature : OK')
            # Strip last 2 signature bytes and return packet
            return packet[:-2]

    def _read_one_byte(self):
        '''Read only one byte.'''
        data = self.link.read(1)
        if is_text(data):
            return bytes(data.encode('utf-8'))
        else:
            return data

    def wait_packet(self, transac_id):
        '''Wait for an incoming packet.

        :param transac_id: Expected transaction number.
        '''

        LOGGER.info('Wait packet with transaction %s' % transac_id)
        data = self.read()
        if data is None or data == b'':
            return {}, {}

        hdr, msg = self.decode_packet(data)
        if hdr == {} or msg == {}:
            return hdr, msg

        # ignore packets that are not for us
        if (hdr['DstNodeId'] != self.src_node) or \
           (hdr['SrcNodeId'] != self.dest_node):
            return {}, {}

        # Respond to incoming hello command packets
        if msg['MsgType'] == 0x09:
            pkt = self.get_hello_response(hdr['SrcNodeId'], hdr['DstNodeId'],
                                          msg['TranNbr'])
            self.write(pkt)
            return self.wait_packet(transac_id)

        # Handle 'please wait' packets
        if msg['TranNbr'] == transac_id and msg['MsgType'] == 0xa1:
            timewait = msg['WaitSec']
            LOGGER.info('Please Wait Message packet <%s sec>' % timewait)
            time.sleep(timewait)
            return self.wait_packet(transac_id)

        # Handle failure message packets and raise exception
        if msg['MsgType'] == 0x81:
            raise DeliveryFailureException()

        # This should be the packet we are waiting for
        if msg['TranNbr'] == transac_id:
            return hdr, msg

    def pack_header(self, hi_proto, exp_more=0x2, link_state=None,
                    hops=0x0):
        '''Generate PakBus header.

        :param hi_proto: Higher level protocol code (4 bits). 0x0: PakCtrl,
                         0x1: BMP5
        :param exp_more: Whether client should expect another packet (2 bits)
        :param link_state: Link state (4 bits)
        :param hops: Number of hops to destination (4 bits)
        '''
        LOGGER.info('Create header')
        priority = 0x1
        link_state = link_state or self.READY
        # bitwise encoding of header fields
        hdr = struct.pack(str('>4H'),
                          (link_state & 0xF) << 12 | (self.dest_node & 0xFFF),
                          (exp_more & 0x3) << 14 | (priority & 0x3) << 12
                                                 | (self.src_node & 0xFFF),
                          (hi_proto & 0xF) << 12 | (self.dest_node & 0xFFF),
                          (hops & 0xF) << 12 | (self.src_node & 0xFFF))
        return hdr

    def compute_signature(self, buff, seed=0xAAAA):
        '''Compute signature for PakBus packets.'''
        sig = seed
        for x in buff:
            x = ord(x)
            j = sig
            sig = (sig << 1) & 0x1FF
            if sig >= 0x100:
                sig += 1
            sig = ((((sig + (j >> 8) + x) & 0xFF) | (j << 8))) & 0xFFFF
        return sig

    def compute_signature_nullifier(self, sig):
        '''Compute signature nullifier needed to create valid PakBus
        packets.'''
        nulb = nullif = b''
        for i in (1, 2):
            sig = self.compute_signature(nulb, sig)
            sig2 = (sig << 1) & 0x1FF
            if sig2 >= 0x100:
                sig2 += 1
            nulb = chr((0x100 - (sig2 + (sig >> 8))) & 0xFF)
            nullif += nulb
        return nullif

    def quote(self, packet):
        '''Quote the PakBus packet.'''
        LOGGER.info('Quote packet')
        packet = packet.replace(b'\xBC', b'\xBC\xDC')
        packet = packet.replace(b'\xBD', b'\xBC\xDD')
        return packet

    def unquote(self, packet):
        '''Unquote the PakBus packet.'''
        LOGGER.info('Unquote packet')
        packet = packet.replace(b'\xBC\xDD', b'\xBD')
        packet = packet.replace(b'\xBC\xDC', b'\xBC')
        return packet

    def encode_bin(self, types, values):
        '''Encode binary data according to data type.'''
        buff = []  # buffer for binary data
        for i, type_ in enumerate(types):
            fmt = self.DATATYPE[type_]['fmt']  # get default format for type_
            value = values[i]

            if type_ == 'ASCIIZ':
                # special handling: nul-terminated string
                value += '\0'
                # Add nul to end of string
                fmt_ = str('%d%s' % (len(value), fmt))
                if is_py3:
                    enc = struct.pack(fmt_, bytes(value, encoding='utf8'))
                else:
                    enc = struct.pack(fmt_, str(value))
            elif type_ == 'ASCII':
                # special handling: fixed-length string
                fmt_ = str('%d%s' % (len(value), fmt))
                enc = struct.pack(fmt_, str(value))
            elif type_ == 'NSec':
                # special handling: NSec time
                enc = struct.pack(str(fmt), value[0], value[1])
            else:
                # default encoding scheme
                enc = struct.pack(str(fmt), value)

            buff.append(enc)
        return b''.join(buff)

    def decode_bin(self, types, buff, length=1):
        '''Decode binary data according to data type.'''
        offset = 0  # offset into buffer
        values = []  # list of values to return
        for type_ in types:
            # get default format and size for Type
            fmt = self.DATATYPE[type_]['fmt']
            size = self.DATATYPE[type_]['size']

            if type_ == 'ASCIIZ':
                # special handling: nul-terminated string
                nul = buff.index(b'\0', offset)
                # find first '\0' after offset
                if nul == -1:
                    value = buff[offset:]
                else:
                    value = buff[offset:nul]
                # return string without trailing '\0'
                size = len(value) + 1
            elif type_ == 'ASCII':
                # special handling: fixed-length string
                size = length
                value = buff[offset:offset + size]
                # return fixed-length string
            elif type_ == 'FP2':
                # special handling: FP2 floating point number
                fp2 = struct.unpack(str(fmt), buff[offset:offset + size])
                mant = fp2[0] & 0x1FFF    # mantissa is in bits 1-13
                exp = fp2[0] >> 13 & 0x3  # exponent is in bits 14-15
                sign = fp2[0] >> 15       # sign is in bit 16
                value = ((-1) ** sign * float(mant) / 10 ** exp, )
            else:
                # default decoding scheme
                if buff[offset:offset + size]:
                    value = struct.unpack(str(fmt), buff[offset:offset + size])
                else:
                    value = ''

            # un-tuple single values
            if len(value) == 1:
                value = value[0]

            values.append(value)
            offset += size
        # Return decoded values and current offset into buffer (size)
        return values, offset

    def decode_packet(self, data):
        '''Decode packet from raw data.'''
        LOGGER.info('Decode packet')
        # pkt: buffer containing unquoted packet, signature nullifier stripped
        # Initialize output variables
        hdr = {'LinkState': None, 'DstPhyAddr': None, 'ExpMoreCode': None,
               'Priority': None, 'SrcPhyAddr': None, 'HiProtoCode': None,
               'DstNodeId': None, 'HopCnt': None, 'SrcNodeId': None}
        msg = {'MsgType': None, 'TranNbr': None, 'raw': None}

        # decode PakBus header
        rawhdr = struct.unpack(str('>4H'), data[0:8])  # raw header bits
        hdr['LinkState'] = rawhdr[0] >> 12
        hdr['DstPhyAddr'] = rawhdr[0] & 0x0FFF
        hdr['ExpMoreCode'] = (rawhdr[1] & 0xC000) >> 14
        hdr['Priority'] = (rawhdr[1] & 0x3000) >> 12
        hdr['SrcPhyAddr'] = rawhdr[1] & 0x0FFF
        hdr['HiProtoCode'] = rawhdr[2] >> 12
        hdr['DstNodeId'] = rawhdr[2] & 0x0FFF
        hdr['HopCnt'] = rawhdr[3] >> 12
        hdr['SrcNodeId'] = rawhdr[3] & 0x0FFF

        # decode default message fields:
        # raw message, message type and transaction number
        msg['raw'] = data[8:]
        values, size = self.decode_bin(('Byte', 'Byte'), msg['raw'][:2])
        msg['MsgType'], msg['TranNbr'] = values
        LOGGER.info('HiProtoCode, MsgType = <%x, %x>' %
                    (hdr['HiProtoCode'], msg['MsgType']))

        # PakBus Control Packets
        if hdr['HiProtoCode'] == 0 and msg['MsgType'] in (0x09, 0x89):
            msg = self.unpack_hello_response(msg)
        elif hdr['HiProtoCode'] == 0 and msg['MsgType'] == 0x81:
            msg = self.unpack_failure_response(msg)
        elif hdr['HiProtoCode'] == 0 and msg['MsgType'] == 0x8f:
            msg = self.unpack_getsettings_response(msg)
        # BMP5 Application Packets
        elif hdr['HiProtoCode'] == 1 and msg['MsgType'] == 0x89:
            msg = self.unpack_collectdata_response(msg)
        elif hdr['HiProtoCode'] == 1 and msg['MsgType'] == 0x97:
            msg = self.unpack_clock_response(msg)
        elif hdr['HiProtoCode'] == 1 and msg['MsgType'] == 0x98:
            msg = self.unpack_getprogstat_response(msg)
        elif hdr['HiProtoCode'] == 1 and msg['MsgType'] == 0x9d:
            msg = self.unpack_fileupload_response(msg)
        elif hdr['HiProtoCode'] == 1 and msg['MsgType'] == 0xa1:
            msg = self.unpack_pleasewait_response(msg)
        else:
            LOGGER.error('No implementation for <(%r, %r)> packet type'
                         % (hdr['HiProtoCode'], msg['MsgType']))
        return hdr, msg

    def get_hello_cmd(self):
        '''Create Hello Command packet.'''
        transac_id = self.transaction.next_id()
        hdr = self.pack_header(0x0, 0x1, self.RING)
        msg = self.encode_bin(['Byte', 'Byte', 'Byte', 'Byte', 'UInt2'],
                              [0x09, transac_id, 0x00, 0x02, 1800])
        return b''.join((hdr, msg)), transac_id

    def get_hello_response(self, transac_id):
        '''Create Hello Response packet.'''
        hdr = self.pack_header(0x0)
        msg = self.encode_bin(['Byte', 'Byte', 'Byte', 'Byte', 'UInt2'],
                              [0x89, transac_id, 0x00, 0x02, 1800])
        return b''.join([hdr, msg])

    def unpack_hello_response(self, msg):
        '''Create Hello Response packet.'''
        raw = msg['raw'][2:]
        values, size = self.decode_bin(['Byte', 'Byte', 'UInt2'], raw)
        msg['IsRouter'], msg['HopMetric'], msg['VerifyIntv'] = values
        return msg

    def unpack_failure_response(self, msg):
        '''Unpack Failure Response packet.'''
        (msg['ErrCode'],), size = self.decode_bin(['Byte'], msg['raw'][2:])
        return msg

    def get_getsettings_cmd(self):
        '''Create Getsettings Command packet.'''
        transac_id = self.transaction.next_id()
        hdr = self.pack_header(0x0)
        msg = self.encode_bin(['Byte', 'Byte'], [0x0f, transac_id])
        return b''.join([hdr, msg]), transac_id

    def unpack_getsettings_response(self, msg):
        '''Unpack Getsettings Response packet.'''
        (msg['Outcome'],), size = self.decode_bin(['Byte'], msg['raw'][2:])
        offset = size + 2

        # Generate dictionary of all settings
        msg['Settings'] = []
        if msg['Outcome'] == 0x01:
            values, size = self.decode_bin(['UInt2', 'Byte', 'Byte', 'Byte'],
                                           msg['raw'][offset:])
            msg['DeviceType'] = values[0]
            msg['MajorVersion'] = values[1]
            msg['MinorVersion'] = values[2]
            msg['MoreSettings'] = values[3]
            offset += size

            while offset < len(msg['raw']):
                # Get setting ID
                [SettingId], size = self.decode_bin(['UInt2'],
                                                    msg['raw'][offset:])
                offset += size
                if not msg['raw'][offset:]:
                    break
                # Get flags and length
                [bit16], size = self.decode_bin(['UInt2'], msg['raw'][offset:])
                LargeValue = (bit16 & 0x8000) >> 15
                ReadOnly = (bit16 & 0x4000) >> 14
                SettingLen = bit16 & 0x3FFF
                offset += size

                # Get value
                SettingValue = msg['raw'][offset:offset + SettingLen]
                offset += SettingLen
                item = {'SettingId': SettingId,
                        'SettingValue': SettingValue,
                        'LargeValue': LargeValue,
                        'ReadOnly': ReadOnly}
                msg['Settings'].append(item)
        return msg

    def get_collectdata_cmd(self, tablenbr, tabledefsig, mode=0x04, p1=0,
                            p2=0):
        '''Create Collect Data Command packet

        :param tablenbr: Table number that contain data.
        :param tabledefsig: Table definition signature.
        :param mode: Collection mode code (p1 and p2 will be used depending on
                    value).
        :param p1: 1st parameter used to specify what to collect (optional)
        :param p2: 2nd parameter used to specify what to collect (optional)
        '''
        transac_id = self.transaction.next_id()
        # BMP5 Application Packet
        hdr = self.pack_header(0x1)
        msg = self.encode_bin(['Byte', 'Byte', 'UInt2', 'Byte'],
                              [0x09, transac_id, self.security_code,
                               mode])
        # encode table number and signature
        msg += self.encode_bin(['UInt2', 'UInt2'], [tablenbr, tabledefsig])
        # add p1 and p2 according to collect mode
        if (mode == 0x04) | (mode == 0x05):
            # only P1 used (type UInt4)
            msg += self.encode_bin(['UInt4'], [p1])
        elif (mode == 0x06) | (mode == 0x08):
            # P1 and P2 used (type UInt4)
            msg += self.encode_bin(['UInt4', 'UInt4'], [p1, p2])
        elif mode == 0x07:
            # P1 and P2 used (type NSec)
            msg += self.encode_bin(['NSec', 'NSec'], [p1, p2])
        # add field list = all fields
        msg += self.encode_bin(['UInt2'], [0])
        return b''.join((hdr, msg)), transac_id

    def unpack_collectdata_response(self, msg):
        '''Unpack Collect Data Response body.'''
        (msg['RespCode'],), size = self.decode_bin(['Byte'], msg['raw'][2:])
        # return raw record data for later parsing
        msg['RecData'] = msg['raw'][size + 2:]
        return msg

    def get_clock_cmd(self, adjustment=(0, 0)):
        '''Create Clock Command packet.

        :param adjustment: Clock adjustment (seconds, nanoseconds).
        '''
        transac_id = self.transaction.next_id()
        # BMP5 Application Packet
        hdr = self.pack_header(0x1)
        msg = self.encode_bin(['Byte', 'Byte', 'UInt2', 'NSec'],
                              [0x17, transac_id, self.security_code,
                               adjustment])
        return b''.join((hdr, msg)), transac_id

    def unpack_clock_response(self, msg):
        '''Unpack Clock Response packet.'''
        values, size = self.decode_bin(['Byte', 'NSec'],  msg['raw'][2:])
        msg['RespCode'], msg['Time'] = values
        return msg

    def get_getprogstat_cmd(self):
        '''Create Get Programming Statistics Transaction packet.'''
        transac_id = self.transaction.next_id()
        # BMP5 Application Packet
        hdr = self.pack_header(0x1)
        msg = self.encode_bin(['Byte', 'Byte', 'UInt2'],
                              [0x18, transac_id, self.security_code])
        return b''.join((hdr, msg)), transac_id

    def unpack_getprogstat_response(self, msg):
        '''Unpack Get Programming Statistics Response packet.'''
        # Get response code
        (msg['RespCode'], ), size = self.decode_bin(['Byte'], msg['raw'][2:])

        # Get report data if RespCode == 0
        if msg['RespCode'] == 0:
            types = ['ASCIIZ', 'UInt2', 'ASCIIZ', 'ASCIIZ', 'Byte', 'ASCIIZ',
                     'UInt2', 'NSec', 'ASCIIZ']
            values, size = self.decode_bin(types, msg['raw'][3:])
            item = {'OSVer': values[0], 'OSSig': values[1],
                    'SerialNbr': values[2], 'PowUpProg': values[3],
                    'CompState': values[4], 'ProgName': values[5],
                    'ProgSig': values[6], 'CompTime': values[7],
                    'CompResult': values[8]}
            msg['Stats'] = item
        return msg

    def get_fileupload_cmd(self, filename, offset=0x00000000, swath=0x0200,
                           closeflag=0x01, transac_id=None):
        '''Create Fileupload Command packet.

        :param filename: File name as string
        :param offset: Byte offset into the file or fragment
        :param swath: Number of bytes to read
        :param closeflag: Flag if file should be closed after this transaction
        :param transac_id: Transaction number for continuing partial reads
                           (required by OS>=17!)
        '''
        if  transac_id is None:
            transac_id = self.transaction.next_id()
        # BMP5 Application Packet
        hdr = self.pack_header(0x1)
        types = ['Byte', 'Byte', 'UInt2', 'ASCIIZ', 'Byte', 'UInt4', 'UInt2']
        values = [0x1d, transac_id, self.security_code, filename, closeflag,
                  offset, swath]
        msg = self.encode_bin(types, values)
        return b''.join((hdr, msg)), transac_id

    def unpack_fileupload_response(self, msg):
        '''Unpack Fileupload Response packet.'''
        values, size = self.decode_bin(['Byte', 'UInt4'], msg['raw'][2:7])
        msg['RespCode'], msg['FileOffset'] = values
        msg['FileData'] = msg['raw'][7:]
        return msg

    def unpack_pleasewait_response(self, msg):
        '''Unpack PeaseWait Response packet.'''
        values, size = self.decode_bin(['Byte', 'UInt2'], msg['raw'][2:])
        msg['CmdMsgType'], msg['WaitSec'] = values
        return msg

    def get_bye_cmd(self):
        '''Create Bye Command packet.'''
        transac_id = self.transaction.next_id()
        # PakBus Control Packet
        hdr = self.pack_header(0x0, 0x0, self.FINISHED)
        msg = self.encode_bin(['Byte', 'Byte'], [0x0d, 0x0])
        return b''.join((hdr, msg)), transac_id

    def parse_filedir(self, data):
        '''Parse file directory format.'''
        offset = 0  # offset into raw buffer
        fd = {'files': []}  # initialize file directory structure
        [fd['DirVersion']], size = self.decode_bin(['Byte'], data[offset:])
        offset += size
        # Extract file entries
        while len(data) > offset:
            file_ = {}  # file description
            [filename], size = self.decode_bin(['ASCIIZ'], data[offset:])
            offset += size

            # end loop when file attribute list terminator reached
            if filename == '':
                break

            file_['FileName'] = filename
            values, size = self.decode_bin(['UInt4', 'ASCIIZ'], data[offset:])
            file_['FileSize'], file_['LastUpdate'] = values
            offset += size

            # Read file attribute list
            file_['Attribute'] = []
            # initialize file attribute list (up to 12)
            for i in range(12):
                [attribute], size = self.decode_bin(['Byte'], data[offset:])
                offset += size
                if attribute:
                    # append file attribute to list
                    file_['Attribute'].append(attribute)
                else:
                    break  # End of attribute list reached
            fd['files'].append(file_)  # add file entry to list
        return fd

    def parse_tabledef(self, raw):
        '''Parse table definition.'''
        tabledef = []  # List of table definitions
        offset = 0  # offset into raw buffer
        fslversion, size = self.decode_bin(['Byte'], raw[offset:])
        offset += size

        # Parse list of table definitions
        while offset < len(raw):

            tblhdr = {}  # table header
            tblfld = []  # table field definitions
            start = offset  # start of table definition

            # Extract table header data
            types = ['ASCIIZ', 'UInt4', 'Byte', 'NSec', 'NSec']
            values, size = self.decode_bin(types, raw[offset:])
            tblhdr['TableName'] = values[0]
            tblhdr['TableSize'] = values[1]
            tblhdr['TimeType'] = values[2]
            tblhdr['TblTimeInto'] = values[3]
            tblhdr['TblInterval'] = values[4]

            offset += size

            # Extract field definitions
            (fieldtype,), size = self.decode_bin(['Byte'], raw[offset:])
            offset += size
            while fieldtype != 0:
                fld = {}

                # Extract bits from fieldtype
                fld['ReadOnly'] = fieldtype >> 7  # only Bit 7

                # Convert fieldtype to ASCII FieldType (e.g. 'FP4') if possible
                # else return numerical value
                fld['FieldType'] = fieldtype & 0x7F  # only Bits 0..6
                for Type in self.DATATYPE.keys():
                    if fld['FieldType'] == self.DATATYPE[Type]['code']:
                        fld['FieldType'] = Type
                        break

                # Extract field name
                values, size = self.decode_bin(['ASCIIZ'], raw[offset:])
                fld['FieldName'] = values[0]
                offset += size

                # Extract AliasName list
                fld['AliasName'] = []
                aliasname = b'00'
                # Alias names list terminator reached
                while aliasname != b'':
                    values, size = self.decode_bin(['ASCIIZ'], raw[offset:])
                    aliasname = values[0]
                    offset += size
                    if aliasname != b'':
                        fld['AliasName'].append(aliasname)

                # Extract other mandatory field definition items
                types = ['ASCIIZ', 'ASCIIZ', 'ASCIIZ', 'UInt4', 'UInt4']
                values, size = self.decode_bin(types, raw[offset:])
                fld['Processing'] = values[0]
                fld['Units'] = values[1]
                fld['Description'] = values[2]
                fld['BegIdx'] = values[3]
                fld['Dimension'] = values[4]
                offset += size

                # Extract sub dimension (if any)
                fld['SubDim'] = []
                subdim = 1
                # sub-dimension list terminator reached
                while subdim != 0:
                    (subdim,), size = self.decode_bin(['UInt4'], raw[offset:])
                    offset += size
                    if subdim != 0:
                        fld['SubDim'].append(subdim)

                # append current field definition to list
                tblfld.append(fld)

                (fieldtype,), size = self.decode_bin(['Byte'], raw[offset:])
                offset += size
            # calculate table signature
            tblsig = self.compute_signature(raw[start:offset])

            # Append header, field list and signature to table definition list
            item = {'Header': tblhdr, 'Fields': tblfld, 'Signature': tblsig}
            tabledef.append(item)
        return tabledef

    def parse_collectdata(self, raw, tabledef, fieldnbr=[]):
        '''Parse data returned by Collectdata Response.'''
        offset = 0
        recdata = []  # output structure

        while offset < len(raw) - 1:
            frag = {}  # record fragment

            values, size = self.decode_bin(['UInt2', 'UInt4'], raw[offset:])
            frag['TableNbr'], frag['BegRecNbr'] = values
            offset += size

            # Provide table name
            t_frag = tabledef[frag['TableNbr'] - 1]
            tablename = t_frag['Header']['TableName']
            frag['TableName'] = tablename

            # Decode number of records (16 bits) or ByteOffset (32 Bits)
            (isoffset,), size = self.decode_bin(['Byte'], raw[offset:])
            frag['IsOffset'] = isoffset >> 7

            # Handle fragmented records
            if frag['IsOffset']:
                (byteoffset,), size = self.decode_bin(['UInt4'], raw[offset:])
                offset += size
                frag['ByteOffset'] = byteoffset & 0x7FFFFFFF
                frag['NbrOfRecs'] = None
                # Copy remaining raw data into RecFrag
                frag['RecFrag'] = raw[offset:-1]
                offset += len(frag['RecFrag'])

            # Handle complete records (standard case)
            else:
                (nbrofrecs,), size = self.decode_bin(['UInt2'], raw[offset:])
                offset += size
                frag['NbrOfRecs'] = nbrofrecs & 0x7FFF
                frag['ByteOffset'] = None

                # Get time of first record and time interval information
                interval = t_frag['Header']['TblInterval']
                if interval == (0, 0):  # event-driven table
                    timeofrec = None
                else:
                    # interval data, read time of first record
                    [timeofrec], size = self.decode_bin(['NSec'], raw[offset:])
                    offset += size

                # Loop over all records
                frag['RecFrag'] = []
                for n in range(frag['NbrOfRecs']):
                    record = {}

                    # Calculate current record number
                    record['RecNbr'] = frag['BegRecNbr'] + n

                    # Get TimeOfRec for interval data or event-driven tables
                    if timeofrec:  # interval data
                        next_timeofrec = (timeofrec[0] + n * interval[0],
                                          timeofrec[1] + n * interval[1])
                        record['TimeOfRec'] = nsec_to_time(next_timeofrec)
                    else:
                        # event-driven, time data precedes each record
                        values, size = self.decode_bin(['NSec'], raw[offset:])
                        record['TimeOfRec'] = values[0]
                        record['TimeOfRec'] = nsec_to_time(record['TimeOfRec'])
                        offset += size

                    # Loop over all field indices
                    record['Fields'] = {}
                    if fieldnbr:
                        # explicit field numbers provided
                        fields = fieldnbr
                    else:
                        # default: generate list of all fields in table
                        fields = t_frag['Fields']
                        fields = range(1, len(fields) + 1)

                    for field in fields:
                        fieldname = t_frag['Fields'][field - 1]['FieldName']
                        fieldtype = t_frag['Fields'][field - 1]['FieldType']
                        dimension = t_frag['Fields'][field - 1]['Dimension']
                        if fieldtype == 'ASCII':
                            values, size = self.decode_bin([fieldtype],
                                                           raw[offset:],
                                                           dimension)
                            record['Fields'][fieldname] = values[0]
                        else:
                            values, size = \
                                self.decode_bin(dimension * [fieldtype],
                                                raw[offset:])
                            record['Fields'][fieldname] = values[0]
                        offset += size
                    frag['RecFrag'].append(record)

            recdata.append(frag)

        # Get flag if more records exist
        (more_rec,), size = self.decode_bin(['Bool'], raw[offset:])
        return recdata, more_rec

    def __del__(self):
        self.link.close()

    def __str__(self):
        name = self.__class__.__name__
        return '<%s %s>' % (name, self.link)

    def __repr__(self):
        return '%s' % self.__str__()



# coding: utf8
'''
    PyCampbellCRX.tests.test_utils
    ------------------------------

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import unicode_literals
import random

from ..utils import (cached_property, Dict, hex_to_bytes, bytes_to_hex,
                     csv_to_dict)
from ..compat import StringIO, is_text, is_bytes


def test_is_text_or_byte():
    '''Tests is text.'''
    assert is_text("Text") is True
    assert is_text(b"\xFF\xFF") is False
    assert is_bytes(b"\xFF\xFF") is True
    assert is_bytes("Text") is False


def test_csv_to_dict():
    '''Tests csv to dict.'''
    file_input = StringIO("a,f\r\n111,222")
    items = csv_to_dict(file_input)
    assert items[0]["a"] == "111"
    assert items[0]["f"] == "222"


def test_dict():
    '''Tests DataDict.'''
    d = Dict()
    d["f"] = "222"
    d["a"] = "111"
    d["b"] = "000"
    assert "a" in d.filter(['a', 'b'])
    assert "b" in d.filter(['a', 'b'])
    assert "f" not in d.filter(['a', 'b'])
    assert "a,f\r\n111,222\r\n" == d.filter(['a', 'f']).to_csv()
    assert "f,b\r\n222,000\r\n" == d.filter(['f', 'b']).to_csv()


def test_ordered_dict():
    '''Tests DataDict.'''
    d = Dict()
    d["f"] = "222"
    d["a"] = "111"
    d["b"] = "000"
    assert "f,a,b\r\n222,111,000\r\n" == d.to_csv()


class TestCachedProperty:
    ''' Tests cached_property decorator.'''

    @cached_property
    def random_bool(self):
        '''Returns random bool'''
        return bool(random.getrandbits(1))

    def test_cached_property(self):
        '''Tests cached_property decorator.'''
        value1 = self.random_bool
        value2 = self.random_bool
        assert value1 == value2


def test_bytes_to_hex():
    '''Tests byte <-> hex and hex <-> byte.'''
    assert bytes_to_hex(b"\xFF") == "FF"
    assert hex_to_bytes(bytes_to_hex(b"\x4A")) == b"\x4A"
    assert bytes_to_hex(hex_to_bytes("4A")) == "4A"



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.tests
    ----------------------

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import unicode_literals

import pytest

from contextlib import contextmanager


@contextmanager
def assert_raises(exception_class, message_part):
    """
    Check that an exception is raised and its message contains some string.
    """
    with pytest.raises(exception_class) as exception:
        yield
    message = '%s' % exception
    assert message_part.lower() in message.lower()



# coding: utf8
'''
    PyCampbellCR1000.compat
    -----------------------

    Workarounds for compatibility with Python 2 and 3 in the same code base.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

#: Python 3.0.x
is_py30 = (is_py3 and _ver[1] == 0)

#: Python 3.1.x
is_py31 = (is_py3 and _ver[1] == 1)

#: Python 3.2.x
is_py32 = (is_py3 and _ver[1] == 2)

#: Python 3.3.x
is_py33 = (is_py3 and _ver[1] == 3)

#: Python 3.4.x
is_py34 = (is_py3 and _ver[1] == 4)

#: Python 2.7.x
is_py27 = (is_py2 and _ver[1] == 7)

#: Python 2.6.x
is_py26 = (is_py2 and _ver[1] == 6)

# ---------
# Specifics
# ---------

if is_py2:
    if is_py26:
        from logging import Handler

        class NullHandler(Handler):
            def emit(self, record):
                pass
        from ordereddict import OrderedDict
    else:
        from logging import NullHandler
        from collections import OrderedDict
    from StringIO import StringIO

    ord = ord
    chr = chr

    def to_char(string):
        if len(string) == 0:
            return bytes('')
        return bytes(string[0])

    bytes = str
    str = unicode
    stdout = sys.stdout
    xrange = xrange


elif is_py3:
    from collections import OrderedDict
    from logging import NullHandler
    from io import StringIO

    ord = lambda x: x
    chr = lambda x: bytes([x])

    def to_char(string):
        if len(string) == 0:
            return str('')
        return str(string[0])

    str = str
    bytes = bytes
    stdout = sys.stdout.buffer
    xrange = range


def is_text(data):
    '''Check if data is text instance'''
    return isinstance(data, str)


def is_bytes(data):
    '''Check if data is bytes instance'''
    return isinstance(data, bytes)



# coding: utf8
'''
    PyCampbellCR1000
    ----------------

    Communication tools for Campbell CR1000-type Dataloggers.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: BSD, see LICENSE for details.

'''
import sys
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

README = ''
CHANGES = ''
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
except:
    pass

REQUIREMENTS = [
    'pylink',
]

if sys.version_info < (2, 7):
    REQUIREMENTS.append('ordereddict')


if sys.version_info < (2, 7) or (3,) <= sys.version_info < (3, 2):
    # In the stdlib from 2.7:
    REQUIREMENTS.append('argparse')

setup(
    name='PyCampbellCR1000',
    version='0.2',
    url='https://github.com/SalemHarrache/PyCampbellCR1000',
    license='GNU GPL v3',
    description='Communication tools for Campbell CR1000-type Dataloggers',
    long_description=README + '\n\n' + CHANGES,
    author='Salem Harrache',
    author_email='salem.harrache@gmail.com',
    maintainer='Lionel Darras',
    maintainer_email='Lionel.Darras@obs.ujf-grenoble.fr',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Topic :: Internet',
        'Topic :: Utilities',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    packages=find_packages(),
    zip_safe=False,
    install_requires=REQUIREMENTS,
    test_suite='pycampbellcr1000.tests',
    entry_points={
        'console_scripts': [
            'pycr1000 = pycampbellcr1000.__main__:main'
        ],
    },
)



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.utils
    -------------------

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import unicode_literals
import math
import time
import calendar
import csv
import binascii

from datetime import datetime

from .compat import to_char, str, StringIO, is_py3, OrderedDict


class Singleton(object):
    '''Signleton class , only one obejct of this type can be created
    any class derived from it will be Singleton.
    Stolen from:
    http://code.activestate.com/recipes/519627-singleton-base-class/'''
    __instance = None

    def __new__(typ, *args, **kwargs):
        if Singleton.__instance is None:
            obj = object.__new__(typ, *args, **kwargs)
            Singleton.__instance = obj

        return Singleton.__instance


class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    Stolen from:
    https://raw.github.com/mitsuhiko/werkzeug/master/werkzeug/utils.py
    """

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        value = obj.__dict__.get(self.__name__)
        if value is None:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def nsec_to_time(nsec, utc=True):
    '''Convert nsec to datetime.'''
    nsec_base = calendar.timegm((1990, 1, 1, 0, 0, 0))
    nsec_tick = 1E-9
    timestamp = nsec_base + nsec[0]
    timestamp += nsec[1] * nsec_tick
    if utc:
        return datetime.utcfromtimestamp(timestamp).replace(microsecond=0)
    return datetime.fromtimestamp(timestamp).replace(microsecond=0)


def time_to_nsec(dtime, utc=True):
    '''Convert timestamp to nsec value.'''
    nsec_base = calendar.timegm((1990, 1, 1, 0, 0, 0))
    nsec_tick = 1E-9
    if utc:
        timestamp = calendar.timegm(dtime.utctimetuple())
    else:
        timestamp = calendar.timegm(dtime.timetuple())
    # separate fractional and integer part of timestamp
    (fp, ip) = math.modf(timestamp)
    # Calculate two integer values for NSec
    nsec = (int(ip - nsec_base), int(fp / nsec_tick))
    return nsec


def bytes_to_hex(byte):
    '''Convert a bytearray to it's hex string representation.'''
    if is_py3:
        hexstr = binascii.hexlify(byte).decode("utf-8")
    else:
        hexstr = str(binascii.hexlify(byte))
    data = []
    for i in range(0, len(hexstr), 2):
        data.append("%s" % hexstr[i:i + 2].upper())
    return ' '.join(data)


def hex_to_bytes(hexstr):
    '''Convert a string hex byte values into a byte string.'''
    return binascii.unhexlify(hexstr.replace(' ', '').encode('utf-8'))


def csv_to_dict(file_input, delimiter=','):
    '''Deserialize csv to list of dictionaries.'''
    delimiter = to_char(delimiter)
    table = []
    reader = csv.DictReader(file_input, delimiter=delimiter,
                            skipinitialspace=True)
    for d in reader:
        table.append(d)
    return ListDict(table)


def dict_to_csv(items, delimiter, header):
    '''Serialize list of dictionaries to csv.'''
    content = ""
    if len(items) > 0:
        delimiter = to_char(delimiter)
        output = StringIO()
        csvwriter = csv.DictWriter(output, fieldnames=items[0].keys(),
                                   delimiter=delimiter)
        if header:
            csvwriter.writerow(dict((key, key) for key in items[0].keys()))
            # writeheader is not supported in python2.6
            # csvwriter.writeheader()
        for item in items:
            csvwriter.writerow(dict(item))

        content = output.getvalue()
        output.close()
    return content


class Dict(OrderedDict):
    '''A dict with somes additional methods.'''

    def filter(self, keys):
        '''Create a dict with only the following `keys`.

        >>> mydict = Dict({"name":"foo", "firstname":"bar", "age":1})
        >>> mydict.filter(['age', 'name'])
        {'age': 1, 'name': 'foo'}
        '''
        data = Dict()
        real_keys = set(self.keys()) - set(set(self.keys()) - set(keys))
        for key in keys:
            if key in real_keys:
                data[key] = self[key]
        return data

    def to_csv(self, delimiter=',', header=True):
        '''Serialize list of dictionaries to csv.'''
        return dict_to_csv([self], delimiter, header)


class ListDict(list):
    '''List of dicts with somes additional methods.'''

    def to_csv(self, delimiter=',', header=True):
        '''Serialize list of dictionaries to csv.'''
        return dict_to_csv(list(self), delimiter, header)

    def filter(self, keys):
        '''Create a list of dictionaries with only the following `keys`.

        >>> mylist = ListDict([{"name":"foo", "age":31},
        ...                    {"name":"bar", "age":24}])
        >>> mylist.filter(['name'])
        [{'name': 'foo'}, {'name': 'bar'}]
        '''
        items = ListDict()
        for item in self:
            items.append(item.filter(keys))
        return items

    def sorted_by(self, keyword, reverse=False):
        '''Returns list sorted by `keyword`.'''
        key_ = keyword
        return ListDict(sorted(self, key=lambda k: k[key_], reverse=reverse))



# -*- coding: utf-8 -*-
'''
    PyCampbellCR1000.client
    -----------------------

    Allows data query of Campbell CR1000-type devices.

    :copyright: Copyright 2012 Salem Harrache and contributors, see AUTHORS.
    :license: GNU GPL v3.

'''
from __future__ import division, unicode_literals
import time

from datetime import datetime, timedelta
from pylink import link_from_url

from .logger import LOGGER
from .pakbus import PakBus
from .exceptions import NoDeviceException
from .compat import xrange, is_py3
from .utils import cached_property, ListDict, Dict, nsec_to_time, time_to_nsec


class CR1000(object):
    '''Communicates with the datalogger by sending commands, reads the binary
    data and parses it into usable scalar values.

    :param link: A `PyLink` connection.
    :param dest_node: Destination node ID (12-bit int) (default 0x001)
    :param src_node: Source node ID (12-bit int) (default 0x802)
    :param security_code: 16-bit security code (default 0x0000)
    '''
    connected = False

    def __init__(self, link, dest_node=0x001, src_node=0x802,
                 security_code=0x0000):
        link.open()
        LOGGER.info("init client")
        self.pakbus = PakBus(link, dest_node, src_node, security_code)
        # try ping the datalogger
        for i in xrange(3):
            try:
                if self.ping_node():
                    self.connected = True
                    break
            except NoDeviceException:
                self.pakbus.link.close()
                self.pakbus.link.open()
        if not self.connected:
            raise NoDeviceException()

    @classmethod
    def from_url(cls, url, timeout=10, dest_node=0x001, src_node=0x802,
                 security_code=0x0000):
        ''' Get device from url.

        :param url: A `PyLink` connection URL.
        :param timeout: Set a read timeout value.
        :param dest_node: Destination node ID (12-bit int) (default 0x001)
        :param src_node: Source node ID (12-bit int) (default 0x802)
        :param security_code: 16-bit security code (default 0x0000)
        '''
        link = link_from_url(url)
        link.settimeout(timeout)
        return cls(link, dest_node, src_node)

    def send_wait(self, cmd):
        '''Send command and wait for response packet.'''
        packet, transac_id = cmd
        begin = time.time()
        self.pakbus.write(packet)
        # wait response packet
        response = self.pakbus.wait_packet(transac_id)
        end = time.time()
        send_time = timedelta(seconds=int((end - begin) / 2))
        return response[0], response[1], send_time

    def ping_node(self):
        '''Check if remote host is available.'''
        # send hello command and wait for response packet
        hdr, msg, send_time = self.send_wait(self.pakbus.get_hello_cmd())
        if not (hdr and msg):
            raise NoDeviceException()
        return True

    def gettime(self):
        '''Return the current datetime.'''
        self.ping_node()
        LOGGER.info('Try gettime')
        # send clock command and wait for response packet
        hdr, msg, send_time = self.send_wait(self.pakbus.get_clock_cmd())
        # remove transmission time
        return nsec_to_time(msg['Time']) - send_time

    def settime(self, dtime):
        '''Sets the given `dtime` and returns the new current datetime'''
        LOGGER.info('Try settime')
        current_time = self.gettime()
        self.ping_node()
        diff = dtime - current_time
        diff = diff.days * 86400 + diff.seconds
        # settime (OldTime in response)
        hdr, msg, sdt1 = self.send_wait(self.pakbus.get_clock_cmd((diff, 0)))
        # gettime (NewTime in response)
        hdr, msg, sdt2 = self.send_wait(self.pakbus.get_clock_cmd())
        # remove transmission time

        return nsec_to_time(msg['Time']) - (sdt1 + sdt2)

    @cached_property
    def settings(self):
        '''Get device settings as ListDict'''
        LOGGER.info('Try get settings')
        self.ping_node()
        # send getsettings command and wait for response packet
        hdr, msg, send_time = self.send_wait(self.pakbus.get_getsettings_cmd())
        # remove transmission time
        settings = ListDict()
        for item in msg["Settings"]:
            settings.append(Dict(dict(item)))
        return settings

    def getfile(self, filename):
        '''Get the file content from the datalogger.'''
        LOGGER.info('Try get file')
        self.ping_node()
        data = []
        # Send file upload command packets until no more data is returned
        offset = 0x00000000
        transac_id = None
        while True:
            # Upload chunk from file starting at offset
            cmd = self.pakbus.get_fileupload_cmd(filename,
                                                 offset=offset,
                                                 closeflag=0x00,
                                                 transac_id=transac_id)
            transac_id = cmd[1]
            hdr, msg, send_time = self.send_wait(cmd)
            try:
                if msg['RespCode'] == 1:
                    raise ValueError("Permission denied")
                # End loop if no more data is returned
                if not msg['FileData']:
                    break
                # Append file data
                data.append(msg['FileData'])
                offset += len(msg['FileData'])
            except KeyError:
                break
        return b"".join(data)

    def list_files(self):
        '''List the files available in the datalogger.'''
        data = self.getfile('.DIR')
        # List files in directory
        filedir = self.pakbus.parse_filedir(data)
        return [item['FileName'] for item in filedir['files']]

    @cached_property
    def table_def(self):
        '''Return table definition.'''
        data = self.getfile('.TDF')
        # List tables
        tabledef = self.pakbus.parse_tabledef(data)
        return tabledef

    def list_tables(self):
        '''List the tables available in the datalogger.'''
        return [item['Header']['TableName'] for item in self.table_def]

    def _collect_data(self, tablename, start_date, stop_date):
        '''Collect fragment data from `tablename` from `start_date` to
        `stop_date` as ListDict.'''
        LOGGER.info('Send collect_data cmd')
        mode = 0x07  # collect from p1 to p2 (nsec)
        p1 = time_to_nsec(start_date)
        p2 = time_to_nsec(stop_date)

        tabledef = self.table_def
        # Get table number
        tablenbr = None
        if is_py3:
            tablename = bytes(tablename, encoding="utf-8")
        for i, item in enumerate(tabledef):
            if item['Header']['TableName'] == tablename:
                tablenbr = i + 1
                break
        if tablenbr is None:
            raise StandardError('table %s not found' % tablename)
        # Get table definition signature
        tabledefsig = tabledef[tablenbr - 1]['Signature']

        # Send collect data request
        cmd = self.pakbus.get_collectdata_cmd(tablenbr, tabledefsig, mode,
                                              p1, p2)
        hdr, msg, send_time = self.send_wait(cmd)
        more = True
        data, more = self.pakbus.parse_collectdata(msg['RecData'], tabledef)
        # Return parsed record data and flag if more records exist
        return data, more

    def get_data(self, tablename, start_date=None, stop_date=None):
        '''Get all data from `tablename` from `start_date` to `stop_date` as
        ListDict. By default the entire contents of the data will be
        downloaded.

        :param tablename: Table name that contains the data.
        :param start_date: The beginning datetime record.
        :param stop_date: The stopping datetime record.'''
        records = ListDict()
        for items in self.get_data_generator(tablename, start_date, stop_date):
            records.extend(items)
        return records

    def get_data_generator(self, tablename, start_date=None, stop_date=None):
        '''Get all data from `tablename` from `start_date` to `stop_date` as
        generator. The data can be fragmented into multiple packets, this
        generator can return parsed data from each packet before receiving
        the next one.

        :param tablename: Table name that contains the data.
        :param start_date: The beginning datetime record.
        :param stop_date: The stopping datetime record.
        '''
        self.ping_node()
        start_date = start_date or datetime(1990, 1, 1, 0, 0, 1)
        stop_date = stop_date or datetime.now()
        more = True
        while more:
            records = ListDict()
            data, more = self._collect_data(tablename, start_date, stop_date)
            for i, rec in enumerate(data):
                if not rec["NbrOfRecs"]:
                    more = False
                    break
                for j, item in enumerate(rec['RecFrag']):
                    if start_date <= item['TimeOfRec'] <= stop_date:
                        start_date = item['TimeOfRec']
                        # for no duplicate record
                        if more and ((j == (len(rec['RecFrag']) - 1))
                                     and (i == (len(data) - 1))):
                            break
                        new_rec = Dict()
                        new_rec["Datetime"] = item['TimeOfRec']
                        new_rec["RecNbr"] = item['RecNbr']
                        for key in item['Fields']:
                            new_rec["%s" % key] = item['Fields'][key]
                        records.append(new_rec)

            if records:
                records = records.sorted_by('Datetime')
                yield records.sorted_by('Datetime')
            else:
                more = False

    def getprogstat(self):
        '''Get programming statistics as dict.'''
        LOGGER.info('Try get programming statistics')
        self.ping_node()
        hdr, msg, send_time = self.send_wait(self.pakbus.get_getprogstat_cmd())
        # remove transmission time
        data = Dict(dict(msg['Stats']))
        if data:
            data['CompTime'] = nsec_to_time(data['CompTime'])
        return data

    def bye(self):
        '''Send a bye command.'''
        LOGGER.info("Send bye command")
        if self.connected:
            packet, transac_id = self.pakbus.get_bye_cmd()
            self.pakbus.write(packet)
            self.connected = False

    def __del__(self):
        '''Send bye cmd when object is deleted.'''
        self.bye()
